/* Copyright (c) 1996 NEC Corporation.  All rights reserved.                 */
/*                                                                           */
/* The redistribution, use and modification in source or binary forms of     */
/* this software is subject to the conditions set forth in the copyright     */
/* document ("COPYRIGHT") included with this distribution.                   */
#include "socks5p.h"
#include "buffer.h"
#include "addr.h"
#include "protocol.h"
#include "confutil.h"
#include "wrap.h"
#include "conf.h"
#include "share.h"

#define EXITERROR(err)	( perror("Traceroute"), exit((err)) )

static int verbose   = 0;
static int noname    = 0;

#ifdef TROUTEPROG
static int ischild   = 0;
static int stopchild = 1;

static void StopChild() {
    int childstat;

    if ((ischild > 0) && stopchild) {
        kill(ischild, SIGINT);
        stopchild = 0;
    }

    if (ischild > 0) {
	while (wait(&childstat) != ischild);
    	ischild   = 0;
    }
}

static int Popen(const char *name) {
    int pds[2];

    if (pipe(pds)          < 0)         return -1;
    if (dup2(pds[0], STDIN_FILENO) < 0) return -1;
    close(pds[0]);

    if ((ischild = fork()) < 0)         return(-1);
    
    if (ischild == 0) {
        close(STDIN_FILENO);
        if (dup2(pds[1], STDOUT_FILENO) < 0) _exit(1);
        if (dup2(pds[1], STDERR_FILENO) < 0) _exit(1);
        close(pds[1]);

	if (noname) execlp(TROUTEPROG, "traceroute", "-n", (verbose?"-v":name), verbose?name:NULL, NULL);
	else        execlp(TROUTEPROG, "traceroute", (verbose?"-v":name), verbose?name:NULL, NULL);
	
        _exit(1);
    }

    stopchild = 1;
    close(pds[1]);

    return(ischild);
}

static int TraceProxy(S5NetAddr *proxy) {
    int childpid;
    
    if ((childpid = Popen(lsAddr2Ascii(proxy))) < 0) return -1;

    fprintf(stdout, "Traceroute to the proxy server %s\n", lsAddr2Ascii(proxy));

    DataRelay();
    StopChild();
    return(0);
}
#endif

static void SocksTrace(S5NetAddr *dst, S5NetAddr *proxy) {
    u_char flag = (noname?SOCKS5_FLAG_NONAME:0)|(verbose?SOCKS5_FLAG_VERBOSE:0);

#ifdef TROUTEPROG
    if (TraceProxy(proxy) < 0) EXITERROR(-1);
#endif
    if (InitProxy(proxy, dst, SOCKS_TRACER, flag) < 0) EXITERROR(-1);

    DataRelay();
    exit(0);
}

void usage(char *progname) {
    fprintf(stderr, "Usage: %s [-n] [-v] host\n", progname);
    exit(-1);
}

int main(int argc, char *argv[]) {
    char *hostname = NULL, *progname = NULL;
    S5NetAddr rsin, dest, *psin;
    struct hostent *rhp;
    int rval, nproxies;

    for (progname = *argv, argc--, argv++; argc > 0; argc--, argv++) {
	if (**argv == '-') {
	    if      (!strcmp(*argv, "-n")) noname  = 1;
	    else if (!strcmp(*argv, "-v")) verbose = 1;
	    else {
		fprintf(stderr, "%s: Unrecognized option %s\n", progname, *argv);
		usage(progname);
	    }
	} else {
	    if (hostname != NULL) usage(progname);
	    hostname = *argv;
	}
    }

    if (hostname == NULL) usage(progname);
    if (LIBPREFIX2(init)(NULL) != 0) EXITERROR(-1);

    if ((rhp = LIBPREFIX(gethostbyname)(hostname)) == NULL) EXITERROR(-1);

    memset((char *)&rsin, 0, sizeof(S5NetAddr));
    memcpy((char *)&rsin.sin.sin_addr.s_addr, (char *)rhp->h_addr_list[0], rhp->h_length);
    rsin.sin.sin_family = rhp->h_addrtype;
    rsin.sin.sin_port   = htons(0);

    switch (rval = lsHowToConnect(&rsin, SOCK_STREAM, &psin, &nproxies, lsEffUser(), &dest)) {
    case DIRECT:
#ifndef TROUTEPROG
	fprintf(stderr, "Real traceroute unavailable.\n");
	exit(0);
#else
	if (noname) execlp(TROUTEPROG, "traceroute", "-n", (verbose?"-v":hostname), verbose?hostname:NULL, NULL);
	else        execlp(TROUTEPROG, "traceroute", (verbose?"-v":hostname), verbose?hostname:NULL, NULL);

	_exit(1);
#endif
    case SOCKS5_VERSION:
	SocksTrace(&dest, psin);
    default:
	fprintf(stderr, "Incompatible Socks Server version for use with Traceroute: %d", rval);
	EXITERROR(-1);
    }
}
